-----------------------------------------------------------------------------------------------
-- Client Lua Script for TrackThat
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
require "AchievementsLib"
-----------------------------------------------------------------------------------------------
-- TrackThat Module Definition
-----------------------------------------------------------------------------------------------
local TrackThat = {}

local _canTrack = {}
local _pendingCheck = {}
local _tracking = {}
local _currentZone = ''
local _currentHouse = nil
local _deleting = false
local _rareKillsLoaded = false
local _challengesLoaded = false
local _timeOfPreviousKill = 0

-----------------------------------------------------------------------------------------------
local glog = nil;
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------

local sVersion = "1.7.7"
-----------------------------------------------------------------------------------------------
-- Locals
-----------------------------------------------------------------------------------------------
local tOptions = {}

local tDefaultOptions = {
	ShowKillForAchievementOption = true,
	ShowDoneKillForAchievementOption = false,
	
	ShowWorldBossesForAchievementOption = true,
	ShowDoneWorldBossesForAchievementOption = true,

	ShowHarvestableOption = true,
	ShowHarvestableAllOption = false,
	
	ShowLoreOption = true,
	ShowDiscoveryOption = true;
	ShowCollectOption = true;
	ShowTargetOption = true;
	
	ShowQuestsOption = true;
	ShowDailyQuestsOption = true;
	
	ShowChallenges = true;
	ShowChallengesUnknown = true;
	
	DebugOption = false
}

function TrackThat:OnSave(eType)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then return end
	local tSaved = {
		ShowDoneKillForAchievementOption = tOptions.ShowDoneKillForAchievementOption,
		ShowKillForAchievementOption = tOptions.ShowKillForAchievementOption,
		
		ShowWorldBossesForAchievementOption = tOptions.ShowWorldBossesForAchievementOption,
		ShowDoneWorldBossesForAchievementOption = tOptions.ShowDoneWorldBossesForAchievementOption,
		
		ShowHarvestableOption = tOptions.ShowHarvestableOption,
		ShowHarvestableAllOption = tOptions.ShowHarvestableAllOption,
		
		ShowLoreOption = tOptions.ShowLoreOption,
		ShowDiscoveryOption = tOptions.ShowDiscoveryOption;
		ShowCollectOption = tOptions.ShowCollectOption;
		ShowTargetOption = tOptions.ShowTargetOption;
		
		ShowQuestsOption = tOptions.ShowQuestsOption;
		ShowDailyQuestsOption = tOptions.ShowDailyQuestsOption;
		
		ShowChallengesOption = tOptions.ShowChallengesOption;
		ShowChallengesUnknown = tOptions.ShowChallengesUnknown;
		
		DebugOption = tOptions.DebugOption,
		Location = self.wndMain:GetLocation():ToTable()	}
	return tSaved
end

-----------------------------------------------------------------------------------------------
-- OnRestore is not called the first time the addon is loaded.
-- Set defaults in OnDocLoaded (race condition, be careful)
-----------------------------------------------------------------------------------------------
function TrackThat:OnRestore(eType, data)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then
		return
	end

	for sOptionName, optionValue in pairs(data) do
		tOptions[sOptionName] = optionValue
	end
	
	if (self.wndMain ~= nil and tOptions.Location ~= nil) then
		self.wndMain:MoveToLocation(WindowLocation.new(tOptions.Location))
	end
end

-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function TrackThat:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    -- initialize variables here

    return o
end

function TrackThat:Init()
	local bHasConfigureFunction = true
	local strConfigureButtonText = "TrackThat"
	local tDependencies = {
		-- "UnitOrPackageName",
	}
	
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end 

-----------------------------------------------------------------------------------------------
-- TrackThat OnLoad
-----------------------------------------------------------------------------------------------
function TrackThat:OnLoad()
    -- load our form file
	self.xmlDoc = XmlDoc.CreateFromFile("TrackThat.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
	
	-- For logging
	--local GeminiLogging = Apollo.GetPackage("Gemini:Logging-1.2").tPackage
    --glog = GeminiLogging:GetLogger({
    --    level = GeminiLogging.INFO,
    --    pattern = "%d %n %c %l - %m",
    --    appender = "GeminiConsole"
    --})	

	--TrackThat:DebugPrint("load");

	Apollo.RegisterEventHandler("VarChange_ZoneName", "OnZoneChange", self)
	Apollo.RegisterEventHandler("SubZoneChanged", "OnZoneChange", self)
	
	Apollo.RegisterEventHandler("UnitCreated", "OnUnitCreated", self)
	Apollo.RegisterEventHandler("UnitDestroyed", "OnUnitDestroyed", self)
	Apollo.RegisterEventHandler("UnitEnteredCombat", "OnUnitCombat", self)
	Apollo.RegisterEventHandler("UnitActivationTypeChanged", "OnUnitActivationChanged", self)

end
	
-----------------------------------------------------------------------------------------------
-- TrackThat OnDocLoaded
-----------------------------------------------------------------------------------------------
function TrackThat:OnDocLoaded()
	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then	
		self.wndMain = Apollo.LoadForm(self.xmlDoc, "TrackThatForm", nil, self)
		if self.wndMain == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return
		end	
		
		if (tOptions.Location ~= nil) then
			self.wndMain:MoveToLocation(WindowLocation.new(tOptions.Location))
		end
		self.wndTrackedUnits = self.wndMain:FindChild('TrackedUnits')
	
	    self.wndConfig = Apollo.LoadForm(self.xmlDoc, "TrackThatConfigForm", nil, self)
		if self.wndConfig == nil then
			Apollo.AddAddonErrorText(self, "Could not load the config window for some reason.")
			return
		end
		
		self.wndSettingBtn = self.wndMain:FindChild('BtnConfig')
		self.wndMain:SetSizingMinimum(200,100)
	    self.wndMain:Show(true, true);
		
	    self.wndConfig:Show(false, true)
		self.wndConfig:FindChild("Title"):SetText("TrackThat Options v" .. sVersion)

		-- if the xmlDoc is no longer needed, you should set it to nil
		self.xmlDoc = nil

		--Default options  Race condition here, but order does not matter for this.
		for sOptionName, optionValue in pairs(tDefaultOptions) do
			if tOptions[sOptionName] == nil then
				tOptions[sOptionName] = optionValue
			end
		end

		Apollo.RegisterEventHandler("InterfaceMenuListHasLoaded", "OnInterfaceMenuListHasLoaded", self)
		Apollo.RegisterEventHandler("TrackThatInterfaceList", "OnConfigure", self)
		Apollo.RegisterSlashCommand("trt", "OnConfigure", self)
		Apollo.RegisterSlashCommand("TrackThat", "OnConfigure", self)	
		Apollo.RegisterTimerHandler("OneSecTimer", "OnUpdate", self)
	end
end

-----------------------------------------------------------------------------------------------
function TrackThat:OnUpdate()

	--TrackThat:DebugPrint("Update")

	self.previousPlayer = self.player
	self.player = GameLib.GetPlayerUnit()
	if (self:IsInValidState() == false) then
		self.player = nil
	elseif (self.previousPlayer ~= self.player) then
		-- do something ?
	end
	
	if (_rareKillsLoaded == false) then
		TrackThat:LoadKillAchievements();
	end
	
	if (_challengesLoaded == false) then
		TrackThat:LoadChallenges();
		TrackThat:OnZoneChange(nil, nil);
	end
	
	if (_deleting == true or
		_rareKillsLoaded == false or
		_challengesLoaded == false) then
		--TrackThat:DebugPrint("Skip one update")
		do return end
	end
	
	if (_pendingCheck ~= nil) then
		for i, unit in pairs(_pendingCheck) do
			self:CheckForMatch(unit)
		end
		_pendingCheck = nil
	end
	
	local unitName;
	local canShow;
	local isKnown;
	local unitsToDelete = {}
	for i, track in pairs(_tracking) do
		isKnown = false;

		if (track.type == "Challenge") then
			if (track.unit.IsActivated ~= nil) then
				isKnown = true;
				unitName = track.unit:GetName();
			elseif (track.unit.IsComplete ~= nil) then
				unitName = track.unit:GetName();
			else
				unitName = track.unit['strChecklistEntry'];
			end
			if (isKnown == false) then
				canShow = tOptions.ShowChallengesUnknown;
			elseif (track.unit:GetDistance() >= 0 or
				track.unit:GetDistance() <= 250 or
				HousingLib.GetResidence() ~= nil) then
				canShow = tOptions.ShowChallengesOption;
			else
				canShow = false;
				self:DeleteRow(track.unit:GetId());
			end
		else
			unitName = track.unit:GetName();
			canShow = false		
			if (track.type == "Kill") then
				if (tOptions.ShowKillForAchievementOption == true) then
					canShow = tOptions.ShowDoneKillForAchievementOption;
					if (canShow == false) then
						-- only show those not killed yet
						if (_canTrack[_currentZone] ~= nill and 
							_canTrack[_currentZone]['Kill'][unitName] ~= nil) then
							canShow = _canTrack[_currentZone]['Kill'][unitName].done == false;
						end
						if (canShow == false and
							_tracking[track.unit:GetId()] ~= nil) then
							-- We just killed one like that, so we should not track it anymore
							unitsToDelete[track.unit:GetId()] = track.unit;
						end
					end
				end
			elseif (track.type == "Boss") then
				if (tOptions.ShowWorldBossesForAchievementOption == true) then
					canShow = tOptions.ShowDoneWorldBossesForAchievementOption;
					if (canShow == false) then
						-- only show those not killed yet
						if (_canTrack['World Bosses']['Kill'][unitName] ~= nil) then
							canShow = _canTrack['World Bosses']['Kill'][unitName].done == false;
						end
						if (canShow == false and
							_tracking[track.unit:GetId()] ~= nil) then
							-- We just killed one like that, so we should not track it anymore
							unitsToDelete[unit:GetId()] = unit;
						end
					end
				end 
			elseif (track.type == "Harvest") then
				if (tOptions.ShowHarvestableOption == true) then
					canShow = tOptions.ShowHarvestableAllOption;
					if (canShow == false) then
						canShow = track.unit:CanBeHarvestedBy(self.player);
					end
				end
			elseif (track.type == "Lore") then
				canShow = tOptions.ShowLoreOption;
			elseif (track.type == "Discovery") then
				canShow = tOptions.ShowDiscoveryOption;
			elseif (track.type == "Target") then
				canShow = tOptions.ShowTargetOption;
			elseif (track.type == "Collect") then
				canShow = tOptions.ShowCollectOption;
			elseif (track.type == "Quest") then
				canShow = tOptions.ShowQuestsOption;
			elseif (track.type == "DailyQuest") then
				canShow = tOptions.ShowDailyQuestsOption;
			end
		end
		
		if (canShow == true) then
			if (track.type == "Challenge") then
				if (isKnown == false) then
					self:AddRow(unitName, unitName, track.type, 0)
				elseif (_currentHouse ~= nil or isKnown == false) then
					self:AddRow(track.unit:GetId(), unitName, track.type, 0)
				else
					self:AddRow(track.unit:GetId(), unitName, track.type, track.unit:GetDistance())
				end
			else
				self:AddRow(track.unit:GetId(), unitName, track.type, self:DistanceToUnit(track.unit))
			end
		end
	end
		
	local trackedUnitCount = self.wndTrackedUnits:GetRowCount()
	for i=1, trackedUnitCount,1 do
		local id = tonumber(self.wndTrackedUnits:GetCellText(i, 4))
		local distance = nil
		
		if (_tracking[id] == nil) then
		-- investigate why this happens sometime
		else
			local unit = _tracking[id].unit
			if(unit ~= nil) then
				if (_tracking[id].type == "Challenge") then
					if (_currentHouse == nil and isKnown == false) then
						distance = unit:GetDistance();
					end
				else
					-- Remove dead units
					if(unit:IsDead() == false) then
						distance = self:DistanceToUnit(unit)
					else
						unitsToDelete[unit:GetId()] = unit
					end
				end
			end
			
			-- Some quests get nil distances when you zone out
			if(distance ~= nil) then
				self.wndTrackedUnits:SetCellText(i, 3, string.format("%5d m", distance))
			end
		end
	end
	
	_deleting = true;
	for key, unit in pairs(unitsToDelete) do
		self:OnUnitDestroyed(unit)
	end
	_deleting = false;
	
	-- Sort by distance
	self.wndTrackedUnits:SetSortColumn(3, true)
end

-----------------------------------------------------------------------------------------------
-- TrackThat DebugPrint
-- TrackThat function for sending debug messages only when the debug option is set
-----------------------------------------------------------------------------------------------
function TrackThat:DebugPrint(sMessage)
	if not tOptions.DebugOption then
		return
	end
	if (glog == nil) then
		Print(sMessage)
	else
		glog:info(sMessage)
	end
end	

-----------------------------------------------------------------------------------------------

function TrackThat:LoadKillAchievements()
	--TrackThat:DebugPrint("Loading Kill Achievements")

	local achievements = AchievementsLib.GetAchievementsForCategory(4, true, false);
	local zone;
	local name;
	for i, achievement in pairs(achievements) do
		if (achievement:GetCategory() ~= 'Kill') then
			zone = achievement:GetCategory()
			
			if (_canTrack[zone] == nil) then
				_canTrack[zone] = {}
			end
			if (_canTrack[zone]['Kill'] == nil) then
				_canTrack[zone]['Kill'] = {}
			end			
			
			if achievement:IsChecklist() then
				for j, listItem in pairs(achievement:GetChecklistItems()) do
					name = TrackThat:CleanKillName(listItem['strChecklistEntry'])
					
					if (_canTrack[zone]['Kill'][name] == nil) then
						_canTrack[zone]['Kill'][name] = {}
					end
					_canTrack[zone]['Kill'][name]['done'] = listItem['bIsComplete'];
				end
			else
				name = TrackThat:CleanKillName(achievement:GetName())
						
				if (_canTrack[zone]['Kill'][name] == nil) then
					_canTrack[zone]['Kill'][name] = {}
				end
				_canTrack[zone]['Kill'][name]['done'] = achievement:IsComplete();
			end
		end
	end	
	
	--TrackThat:DebugPrint("Loaded Kill Achievements")
	
	if ((_canTrack['Algoroc'] ~= nil and _canTrack['Algoroc']['Kill'] ~= nil) or
		(_canTrack['Ellevar'] ~= nil and _canTrack['Ellevar']['Kill'] ~= nil)) then
		_rareKillsLoaded = true;
	end
end

function TrackThat:LoadChallenges()
	--TrackThat:DebugPrint("Loading Challenges")
	
	local allKnownChallenges  = ChallengesLib.GetActiveChallengeList();
	local allChallenges = AchievementsLib.GetAchievementsForCategory(161, true, false);
	local zon;
	local name;
	local known = {}
	for idx, clgCurrent in pairs(allKnownChallenges) do
		local zone = clgCurrent:GetZoneInfo()["strZoneName"];

		local name = clgCurrent:GetName();
		if (known[zone] == nil) then
			known[zone] = {}
		end
		known[zone][name] = clgCurrent
		
		if (zone == 'Star-Comm Basin') then
			TrackThat:TrackChallenge(clgCurrent, zone, known);
		end
	end
	
	for idx, clgCurrent in pairs(allChallenges) do
		zone = clgCurrent:GetCategory()	
		if (zone ~= "Challenges") then

			if (_canTrack[zone] == nil) then
				_canTrack[zone] = {}
			end
			if (_canTrack[zone]['Challenge'] == nil) then
				_canTrack[zone]['Challenge'] = {}
			end
			if clgCurrent:IsChecklist() then
				for j, listItem in pairs(clgCurrent:GetChecklistItems()) do
					name = listItem['strChecklistEntry']
					
					if (_canTrack[zone]['Challenge'][name] == nil) then
						_canTrack[zone]['Challenge'][name] = {}
					end
	
					_canTrack[zone]['Challenge'][name]['name'] = name;
					_canTrack[zone]['Challenge'][name]['item'] = listItem;
					_canTrack[zone]['Challenge'][name]['known'] = nil;
					if (known[zone] ~= nil and known[zone][name] ~= nil) then
						_canTrack[zone]['Challenge'][name]['known'] = known[zone][name];
					end
				end
			else
				TrackThat:TrackChallenge(clgCurrent, zone, known);
			end
		end
	end
	
	--TrackThat:DebugPrint("Loaded Challenges")
	
	if ((_canTrack['Algoroc'] ~= nil and _canTrack['Algoroc']['Challenge'] ~= nil) or
		(_canTrack['Ellevar'] ~= nil and _canTrack['Ellevar']['Challenge'] ~= nil)) then
		_challengesLoaded = true;
	end
end

function TrackThat:TrackChallenge(clgCurrent, zone, known)
	if (_canTrack[zone] == nil) then
		_canTrack[zone] = {}
	end
	if (_canTrack[zone]['Challenge'] == nil) then
		_canTrack[zone]['Challenge'] = {}
	end
	
	local name = clgCurrent:GetName()
			
	if (_canTrack[zone]['Challenge'][name] == nil) then
		_canTrack[zone]['Challenge'][name] = {}
	end
	_canTrack[zone]['Challenge'][name]['name'] = name;
	_canTrack[zone]['Challenge'][name]['item'] = clgCurrent;
	_canTrack[zone]['Challenge'][name]['known'] = nil;
	if (known[zone] ~= nil and known[zone][name] ~= nil) then
		_canTrack[zone]['Challenge'][name]['known'] = known[zone][name];
	end
end

table.indexOf = function( t, object )
    if "table" == type( t ) then
        for i = 1, #t do
            if object == t[i] then
                return i
            end
        end
        return -1
    else
		error("table.indexOf expects table for first argument, " .. type(t) .. " given")
    end
end

function string.ends(String, End)
   return End=='' or string.sub(String,-string.len(End))==End
end

function TrackThat:CleanKillName(pName)
	pName = pName:gsub("World Boss: ", "")
	pName = pName:gsub("Exotic Executioner: ", "")
	if (string.ends(pName, "!")) then
		pName = pName:gsub("!", "")
	end
	return pName
end

function TrackThat:IsInValidState()
	return (self.player ~= nil 
		and self.player:IsValid() == true
		and self.wndMain ~= nil)
end

-----------------------------------------------------------------------------------------------
-- TrackThatForm Functions
-----------------------------------------------------------------------------------------------
function TrackThat:OnInterfaceMenuListHasLoaded()
	Event_FireGenericEvent("InterfaceMenuList_NewAddOn", "TrackThat", {"TrackThatInterfaceList", "", ""})
end

function TrackThat:OnConfigure()
	self.wndConfig:Show(true, true)
	self.wndConfig:Invoke(true)
end

function TrackThat:OnShow(wndHandler, wndControl)
	self.wndConfig:FindChild("ShowKillForAchievementOption"):SetCheck(tOptions.ShowKillForAchievementOption)
	self.wndConfig:FindChild("ShowDoneKillForAchievementOption"):SetCheck(tOptions.ShowDoneKillForAchievementOption)

	self.wndConfig:FindChild("ShowWorldBossesForAchievementOption"):SetCheck(tOptions.ShowWorldBossesForAchievementOption)
	self.wndConfig:FindChild("ShowDoneWorldBossesForAchievementOption"):SetCheck(tOptions.ShowDoneWorldBossesForAchievementOption)

	self.wndConfig:FindChild("ShowHarvestableOption"):SetCheck(tOptions.ShowHarvestableOption)
	self.wndConfig:FindChild("ShowHarvestableAllOption"):SetCheck(tOptions.ShowHarvestableAllOption)
	
	self.wndConfig:FindChild("ShowLoreOption"):SetCheck(tOptions.ShowLoreOption)
	self.wndConfig:FindChild("ShowDiscoveryOption"):SetCheck(tOptions.ShowDiscoveryOption)
	self.wndConfig:FindChild("ShowCollectOption"):SetCheck(tOptions.ShowCollectOption)
	self.wndConfig:FindChild("ShowTargetOption"):SetCheck(tOptions.ShowTargetOption)

	self.wndConfig:FindChild("ShowQuestsOption"):SetCheck(tOptions.ShowQuestsOption)
	self.wndConfig:FindChild("ShowDailyQuestsOption"):SetCheck(tOptions.ShowDailyQuestsOption)
	
	self.wndConfig:FindChild("ShowChallengesOption"):SetCheck(tOptions.ShowChallengesOption)
	self.wndConfig:FindChild("ShowChallengesUnknown"):SetCheck(tOptions.ShowChallengesUnknown)	
	
	self.wndConfig:FindChild("DebugOption"):SetCheck(tOptions.DebugOption)
end

-- when the OK button is clicked
function TrackThat:OnOK()
	tOptions.ShowDoneKillForAchievementOption= self.wndConfig:FindChild("ShowDoneKillForAchievementOption"):IsChecked()
	tOptions.ShowKillForAchievementOption = self.wndConfig:FindChild("ShowKillForAchievementOption"):IsChecked()

	tOptions.ShowWorldBossesForAchievementOption= self.wndConfig:FindChild("ShowWorldBossesForAchievementOption"):IsChecked()
	tOptions.ShowDoneWorldBossesForAchievementOption= self.wndConfig:FindChild("ShowDoneWorldBossesForAchievementOption"):IsChecked()
	
	tOptions.ShowHarvestableOption= self.wndConfig:FindChild("ShowHarvestableOption"):IsChecked()
	tOptions.ShowHarvestableAllOption= self.wndConfig:FindChild("ShowHarvestableAllOption"):IsChecked()

	tOptions.ShowLoreOption= self.wndConfig:FindChild("ShowLoreOption"):IsChecked()
	tOptions.ShowDiscoveryOption = self.wndConfig:FindChild("ShowDiscoveryOption"):IsChecked()
	tOptions.ShowCollectOption = self.wndConfig:FindChild("ShowCollectOption"):IsChecked()
	tOptions.ShowTargetOption = self.wndConfig:FindChild("ShowTargetOption"):IsChecked()
	
	tOptions.ShowQuestsOption = self.wndConfig:FindChild("ShowQuestsOption"):IsChecked()
	tOptions.ShowDailyQuestsOption = self.wndConfig:FindChild("ShowDailyQuestsOption"):IsChecked()
	
	tOptions.ShowChallengesOption = self.wndConfig:FindChild("ShowChallengesOption"):IsChecked()
	tOptions.ShowChallengesUnknown= self.wndConfig:FindChild("ShowChallengesUnknown"):IsChecked()	
	
	tOptions.DebugOption = self.wndConfig:FindChild("DebugOption"):IsChecked()
	
	-- Clear the grid to get a proper refresh on next Update
	local trackedUnitCount = self.wndTrackedUnits:GetRowCount()
	local unitsToDelete = {}
	-- We only delete the row from the grid, but keep the data
	for i=trackedUnitCount, 1, -1 do
		self.wndTrackedUnits:DeleteRow(i)
	end
	
	self:OnUpdate();

	self.wndConfig:Close() -- hide the window
end

-- when the Cancel button is clicked
function TrackThat:OnCancel()
	self.wndConfig:Close() -- hide the window
end
-----------------------------------------------------------------------------------------------
function string.starts(String,Start)
   return string.sub(String:lower(),1,string.len(Start:lower()))==Start:lower()
end

---------------------------------------------------------------------------------------------------
-- Apollo Events
---------------------------------------------------------------------------------------------------

function TrackThat:OnUnitCreated(unit)
	--TrackThat:DebugPrint("create " .. unit:GetName() .. " - " .. unit:GetType())
	
	if (_rareKillsLoaded == true) then
		self:CheckForMatch(unit)
	else
		_pendingCheck[unit:GetId()] = unit;
	end
	
	--TrackThat:DebugPrint(unit:GetName() .. " - " .. unit:GetType())

end

function TrackThat:OnUnitActivationChanged(unit)	
	local type = unit:GetType();
	if (type == '') then
		return
	end
	local id = unit:GetId();
	local activation = unit:GetActivationState()
	if (_tracking[id] ~= nil and
		(_tracking[id].type == 'Lore' or _tracking[id].type == 'Target' or _tracking[id].type == 'Collect') and
		activation ~= nil and
		activation.Datacube == nil) then
		-- Lore item that was activated and to be removed
		_deleting = true;
		self:OnUnitDestroyed(unit)
		_deleting = false;
	else
		--TrackThat:DebugPrint("Changed " .. unit:GetName() .. " - " .. unit:GetType())
		-- A unit changed maybe we need to track it now
		self:CheckForMatch(unit)
	end
end

function TrackThat:CheckForMatch(unit)
	local type = unit:GetType();
	if (type == '') then
		return
	end
	local unitName = unit:GetName();
	local activation = unit:GetActivationState()
	local unitRewardInfo = unit:GetRewardInfo()
	
	if (activation ~= nil and 
		(activation.QuestNewMain ~= nil or
		 activation.QuestNewTradeskill ~= nil or
		 activation.QuestNew ~= nil or
		 activation.QuestNewRepeatable ~= nil or
		 activation.QuestNewDaily)) then
		if (activation.QuestNewRepeatable ~= nil or
		 activation.QuestNewDaily) then
			self:Track(unit, 'DailyQuest');
		else
			self:Track(unit, 'Quest');
		end
		
	elseif (type == "NonPlayer") then	
		if (activation ~= nil and activation.Discovery ~= nil) then
			self:Track(unit, 'Discovery');
			
		elseif (activation ~= nil and (activation.ScientistScannable ~= nil or activation.ScientistRawScannable ~= nil)) then
			-- Todo later

		elseif (_canTrack[_currentZone] ~= nill and
			_canTrack[_currentZone]['Kill'] ~= nil and
			_canTrack[_currentZone]['Kill'][unitName] ~= nil) then
			self:Track(unit, 'Kill');

		elseif (_canTrack[_currentZone] ~= nill and
			_canTrack[_currentZone]['Kill'] ~= nil and
			_canTrack['World Bosses']['Kill'][unitName] ~= nil) then
			self:Track(unit, 'Boss');

		elseif (unitRewardInfo ~= nil and table.getn(unitRewardInfo) > 0) then
			--TrackThat:DebugPrint("Nonplayer " .. unit:GetName() .. " - " .. unit:GetType())
			--TrackThat:DebugPrint(activation)
			--TrackThat:DebugPrint(unitRewardInfo)
		
			for i, reward in pairs(unitRewardInfo) do
				--TrackThat:DebugPrint(reward.strType)

				if (reward.strType == "Challenge") then
					-- Todo later
				else
					self:Track(unit, 'Target');
				end
			end
		end
		
	elseif (type == "Harvest") then
		self:Track(unit, 'Harvest');
		
	elseif (activation ~= nil and activation.Datacube ~= nil) then
		self:Track(unit, 'Lore');
		
	elseif (activation ~= nil and (activation.ScientistScannable ~= nil or activation.ScientistRawScannable ~= nil)) then
		--TrackThat:DebugPrint("Scientist " .. unit:GetName() .. " - " .. unit:GetType())
		
	elseif (unitRewardInfo ~= nil and table.getn(unitRewardInfo) > 0) then
		--TrackThat:DebugPrint("Target " .. unit:GetName() .. " - " .. unit:GetType())
		--TrackThat:DebugPrint(unitRewardInfo);
		self:Track(unit, 'Target');
		
	elseif (activation ~= nil and activation.Collect ~= nil) then
		self:Track(unit, 'Collect');
				
	else
		--TrackThat:DebugPrint("Untracked " .. unit:GetName() .. " - " .. unit:GetType())
	end
end

function TrackThat:Track(pUnit, pType)
	local id = pUnit:GetId();
	
	if (_tracking[id] == nil) then
		_tracking[id] = {
			unit = pUnit,
			type = pType
		};
	end
end

function TrackThat:TrackNoID(pUnit, pType, pID)
	if (_tracking[pID] == nil) then
		_tracking[pID] = {
			unit = pUnit,
			type = pType
		};
	end
end

function TrackThat:AddRow(pID, pName, pType, pDistance)
	local index = self:GetRowIndexOfId(pID)
		
	if(index == nil) then
		local tRow = self.wndTrackedUnits:AddRow("")
		self.wndTrackedUnits:SetCellText(tRow, 1, pName)
		self.wndTrackedUnits:SetCellText(tRow, 2, pType)
		self.wndTrackedUnits:SetCellText(tRow, 3, string.format("%5d m", pDistance))
		self.wndTrackedUnits:SetCellText(tRow, 4, pID)
	end
end

function TrackThat:GetRowIndexOfId(id)
	local tRowIndex = nil
	if(self.wndTrackedUnits ~= nil) then
		local trackedUnitCount = self.wndTrackedUnits:GetRowCount()
		for i=1, trackedUnitCount,1 do
			local rowId = tonumber(self.wndTrackedUnits:GetCellText(i, 4))
			if (rowId == nil) then
				rowId = self.wndTrackedUnits:GetCellText(i, 4);
			end
			if(rowId == id) then
				tRowIndex = i
				return tRowIndex
			end
		end
	end
	
	return nil
end

function TrackThat:DeleteRow(id)
	local tRowIndex = self:GetRowIndexOfId(id)	
	if(tRowIndex ~= nil) then
		self.wndTrackedUnits:DeleteRow(tRowIndex)
	end
end

function TrackThat:OnUnitDestroyed(unit)
	local id;
	if (unit['strChecklistEntry'] ~= nil) then
		id = unit['strChecklistEntry'];
	else
		id = unit:GetId();
	end
	if (_tracking[id] ~= nil) then
		_tracking[id] = nil;
		self:DeleteRow(id);

		local now = Apollo.GetTickCount();
		if (now - _timeOfPreviousKill > 5000) then
			-- need to refresh zone related data
			TrackThat:LoadKillAchievements();
			_timeOfPreviousKill = now;
		end
	end
end

local _inCombat = false

function TrackThat:OnUnitCombat(unit)
	if (unit:IsThePlayer() == true) then
		local isFighting = unit:IsCasting()
		if (isFighting ~= _inCombat) then
			_inCombat = isFighting;
			if (_inCombat == true) then
				--TrackThat:DebugPrint("OnUnitCombat - start fighting")
			else
				--TrackThat:DebugPrint("OnUnitCombat - Done fighting")
			end
		end
	end
end

function TrackThat:OnZoneChange(oVar, strNewZone)
	local tCurrentZoneInfo = GameLib.GetCurrentZoneMap()
	if (tCurrentZoneInfo == nil or
		_currentZone ~= tCurrentZoneInfo.strName) then
		local newZoneName = tCurrentZoneInfo.strName;
		if (newZoneName == "Housing Skymap") then
			newZoneName = "Housing";
		end
		
		--TrackThat:DebugPrint("Zone changed from " .. _currentZone .. ' to ' .. newZoneName)
		if (_canTrack[newZoneName] == nil and
			newZoneName ~= "Housing") then
			-- User moved to a subzone (cave,...) so we do not change the zone name
			do return end
		end
		_currentZone = newZoneName;
	end

	if (_canTrack[_currentZone] == nil and
		_currentZone ~= "Housing") then
		-- User moved from a subzone to a subzone (cave,...) so we do not change the zone name
		do return end
	end
				
	if (_challengesLoaded == false) then
		do return end
	end	
	
	local house = HousingLib.GetResidence();
	if (house ~= nil) then
		if (house:IsWarplotResidence() == false) then
			local houseName = house:GetPropertyName();
			if (houseName ~= _currentHouse) then
				if (_currentHouse ~= nil) then
					self:CleanChallenges()
				end
				_currentHouse = houseName;
				
				for i=1, 7, 1 do
					local pltPlot = HousingLib.GetPlot(i)
					if (pltPlot ~= nil and pltPlot:GetPlugName() ~= nil) then
				    	local plugItem = HousingLib.GetPlugItem(pltPlot:GetPlugItemId())
				
						if (string.find(plugItem["strTooltip"], 'challenge', 1, true)) then
							local name = string.gsub(plugItem["strTooltip"], "Place this plug to attempt the \"", "")
							name = string.gsub(name, "\" challenge.", "")
							local dummy = {
								strChecklistEntry= name
							}
							self:TrackNoID(dummy, 'Challenge', name);
						end
	
	-- Looks like Housing Challenges have been removed from the Achievements list...
	-- So we will never hit this again...
	-- Keeping it just in case...
				
						--for i, challenge in pairs(_canTrack[_currentZone]['Challenge']) do
						--	if (string.find(plugItem["strTooltip"], challenge['name'], 1, true)) then
						--		self:Track(challenge['item'], 'Challenge');
						--		break;
						--	end
						--end
					end
				end
			end
		else
			self:CleanChallenges()
		end
	else
		self:CleanChallenges()
		
		if (_canTrack[_currentZone]['Challenge'] ~= nil) then
			for i, challenge in pairs(_canTrack[_currentZone]['Challenge']) do
				if (challenge['known'] == nil) then
					self:TrackNoID(challenge['item'], 'Challenge', challenge['name']);
				else
					self:Track(challenge['known'], 'Challenge');
				end
			end
		end
	end
end

function TrackThat:CleanChallenges()
	-- Got to clean the previous ones
	_currentHouse = nil;
	_deleting = true;
	
	for i, track in pairs(_tracking) do
		if (track.type == "Challenge") then
			self:OnUnitDestroyed(track.unit, false);
		end
	end
	
	_deleting = false;
end
		
function TrackThat:DistanceToUnit(unit)
	local posPlayer = self.player:GetPosition()
	local posTarget = unit:GetPosition()
	local nDeltaX = posTarget.x - posPlayer.x
	local nDeltaY = posTarget.y - posPlayer.y
	local nDeltaZ = posTarget.z - posPlayer.z
	
	return math.sqrt(math.pow(nDeltaX, 2) + math.pow(nDeltaY, 2) + math.pow(nDeltaZ, 2))
end

function TrackThat:OnTrackedItemPick(wndControl, wndHandler, iRow, iCol)
	local id = tonumber(self.wndTrackedUnits:GetCellText(iRow, 4))
	
	if (_currentHouse ~= nil) then
		ChallengesLib.ShowHintArrow(_tracking[id].unit:GetId());
	elseif (_tracking[id].type ~= "Challenge" or
		_tracking[id].unit.IsActivated ~= nil or
		_tracking[id].unit.IsComplete ~= nil)  then
		local retrievedEntity = _tracking[id].unit
		if (_tracking[id].type ~= "Challenge") then
			if(retrievedEntity ~= nill) then
				retrievedEntity:ShowHintArrow()
			end
		else
			ChallengesLib.ShowHintArrow(_tracking[id].unit:GetId())
		end
	end
end

-----------------------------------------------------------------------------------------------
-- TrackThat Instance
-----------------------------------------------------------------------------------------------
local TrackThatInst = TrackThat:new()
TrackThatInst:Init()
